@@ -14,7 +14,7 @@ from streamflow.core.data import FileType, LOCAL_RESOURCE, DataLocationType
 from streamflow.core.deployment import Connector
 from streamflow.core.exception import WorkflowExecutionException, WorkflowDefinitionException, \
     UnrecoverableTokenException
-from streamflow.core.utils import get_path_processor, random_name, flatten_list
+from streamflow.core.utils import get_path_processor, random_name, flatten_list, get_tag
 from streamflow.core.workflow import Step, Port, InputPort, Job, Token, Status, TokenProcessor, CommandOutput
 from streamflow.cwl import utils
 from streamflow.cwl.command import CWLCommandOutput
@@ -146,6 +146,7 @@ class LoadListing(Enum):
 
 
 class SecondaryFile(object):
+    __slots__ = ('pattern', 'required')
 
     def __init__(self,
                  pattern: Text,
@@ -153,6 +154,15 @@ class SecondaryFile(object):
         self.pattern: Text = pattern
         self.required: bool = required
 
+    def __eq__(self, other):
+        if not isinstance(other, SecondaryFile):
+            return False
+        else:
+            return self.pattern == other.pattern
+
+    def __hash__(self):
+        return hash(self.pattern)
+
 
 class CWLTokenProcessor(DefaultTokenProcessor):
 
@@ -722,11 +732,17 @@ class CWLTokenProcessor(DefaultTokenProcessor):
 
     async def compute_token(self, job: Job, command_output: CWLCommandOutput) -> Any:
         if command_output.status == Status.SKIPPED:
-            return Token(name=self.port.name, value=None, job=job.name)
-        token_value = await self._get_value_from_command(job, command_output)
-        self._register_data(job, token_value)
-        weight = await self.weight_token(job, token_value)
-        return Token(name=self.port.name, value=token_value, job=job.name, weight=weight)
+            return None
+        else:
+            token_value = await self._get_value_from_command(job, command_output)
+            self._register_data(job, token_value)
+            weight = await self.weight_token(job, token_value)
+            return Token(
+                name=self.port.name,
+                value=token_value,
+                job=job.name,
+                tag=get_tag(job.inputs),
+                weight=weight)
 
     def get_related_resources(self, token: Token) -> Set[Text]:
         if self.port_type in ['File', 'Directory']:
@@ -762,7 +778,13 @@ class CWLTokenProcessor(DefaultTokenProcessor):
         if token.value is None and self.default_value is not None:
             token = token.update(await self._build_token_value(job, self.default_value))
         if self.port_type == 'Any' or self.port_type is None:
-            self.port_type = utils.infer_type_from_token(token.value)
+            if (isinstance(self.port, InputPort) and
+                    isinstance(self.port.dependee.token_processor, CWLTokenProcessor) and
+                    self.port.dependee.token_processor.port_type != 'Any' and
+                    self.port.dependee.token_processor.port_type is not None):
+                self.port_type = self.port.dependee.token_processor.port_type
+            else:
+                self.port_type = utils.infer_type_from_token(token.value)
         if isinstance(token.value, MutableMapping) and token.value.get('class') in ['File', 'Directory']:
             context = self.get_context()
             src_job = context.scheduler.get_job(token.job)
@@ -805,14 +827,36 @@ class CWLMapTokenProcessor(MapTokenProcessor):
 
     def __init__(self,
                  port: Port,
-                 token_processor: TokenProcessor):
+                 token_processor: TokenProcessor,
+                 default_value: Optional[Any] = None,
+                 optional: bool = False):
         super().__init__(port, token_processor)
+        self.default_value: Optional[Any] = default_value
+        self.optional: bool = optional
 
     async def compute_token(self, job: Job, command_output: CommandOutput) -> Token:
         if isinstance(command_output.value, MutableMapping) and self.port.name in command_output.value:
             command_output = command_output.update([{self.port.name: v} for v in command_output.value[self.port.name]])
         return await super().compute_token(job, command_output)
 
+    async def update_token(self, job: Job, token: Token) -> Token:
+        if token.value is None:
+            token = token.update(self.default_value)
+        return await super().update_token(job, token)
+
+
+class CWLSkipTokenProcessor(DefaultTokenProcessor):
+
+    async def compute_token(self, job: Job, command_output: CWLCommandOutput) -> Any:
+        if command_output.status == Status.SKIPPED:
+            return Token(
+                name=self.port.name,
+                value=None,
+                job=job.name,
+                tag=get_tag(job.inputs))
+        else:
+            return None
+
 
 class CWLUnionTokenProcessor(UnionTokenProcessor):
 
