@@ -11,6 +11,7 @@ import tempfile
 import uuid
 from abc import ABC
 from typing import MutableMapping, MutableSequence, Optional, Any, Tuple, Union
+from urllib.parse import urlencode
 
 import yaml
 from cachetools import Cache, TTLCache
@@ -43,6 +44,53 @@ class PatchedInClusterConfigLoader(incluster_config.InClusterConfigLoader):
         configuration.api_key['authorization'] = "bearer " + self.token
 
 
+class PatchedWsApiClient(WsApiClient):
+
+    async def request(self, method, url, query_params=None, headers=None,
+                      post_params=None, body=None, _preload_content=True,
+                      _request_timeout=None):
+        if query_params:
+            new_query_params = []
+            for key, value in query_params:
+                if key == 'command' and isinstance(value, list):
+                    for command in value:
+                        new_query_params.append((key, command))
+                else:
+                    new_query_params.append((key, value))
+            query_params = new_query_params
+        if headers is None:
+            headers = {}
+        if 'sec-websocket-protocol' not in headers:
+            headers['sec-websocket-protocol'] = 'v4.channel.k8s.io'
+
+        if query_params:
+            url += '?' + urlencode(query_params)
+
+        url = ws_client.get_websocket_url(url)
+
+        if _preload_content:
+
+            resp_all = ''
+            async with self.rest_client.pool_manager.ws_connect(
+                    url,
+                    headers=headers,
+                    heartbeat=30) as ws:
+                async for msg in ws:
+                    msg = msg.data.decode('utf-8')
+                    if len(msg) > 1:
+                        channel = ord(msg[0])
+                        data = msg[1:]
+                        if data:
+                            if channel in [ws_client.STDOUT_CHANNEL, ws_client.STDERR_CHANNEL]:
+                                resp_all += data
+
+            return ws_client.WsResponse(resp_all.encode('utf-8'))
+
+        else:
+
+            return await self.rest_client.pool_manager.ws_connect(url, headers=headers, heartbeat=30)
+
+
 class BaseHelmConnector(BaseConnector, ABC):
 
     def __init__(self,
@@ -50,13 +98,12 @@ class BaseHelmConnector(BaseConnector, ABC):
                  inCluster: Optional[bool] = False,
                  kubeconfig: Optional[Text] = os.path.join(os.environ['HOME'], ".kube", "config"),
                  namespace: Optional[Text] = None,
-                 readBufferSize: Optional[int] = None,
                  releaseName: Optional[Text] = "release-%s" % str(uuid.uuid1()),
                  resourcesCacheTTL: int = 10,
-                 transferBufferSize: int = (2 ** 25) - 1):
+                 transferBufferSize: int = (2 ** 25) - 1,
+                 maxConcurrentConnections: int = 4096):
         super().__init__(
             streamflow_config_dir=streamflow_config_dir,
-            readBufferSize=readBufferSize,
             transferBufferSize=transferBufferSize)
         self.inCluster = inCluster
         self.kubeconfig = kubeconfig
@@ -66,6 +113,7 @@ class BaseHelmConnector(BaseConnector, ABC):
         self.configuration: Optional[Configuration] = None
         self.client: Optional[client.CoreV1Api] = None
         self.client_ws: Optional[client.CoreV1Api] = None
+        self.maxConcurrentConnections: int = maxConcurrentConnections
 
     def _configure_incluster_namespace(self):
         if self.namespace is None:
@@ -108,9 +156,9 @@ class BaseHelmConnector(BaseConnector, ABC):
                                            resource: Text,
                                            tar_buffer: io.BufferedRandom,
                                            read_only: bool = False) -> None:
-        resource_buffer = io.BufferedReader(tar_buffer.raw, buffer_size=self.readBufferSize)
+        resource_buffer = io.BufferedReader(tar_buffer.raw)
         pod, container = resource.split(':')
-        command = ['tar', 'xzf', '-', '-C', '/']
+        command = ["tar", "xf", "-", "-C", "/"]
         response = await self.client_ws.connect_get_namespaced_pod_exec(
             name=pod,
             namespace=self.namespace or 'default',
@@ -133,7 +181,7 @@ class BaseHelmConnector(BaseConnector, ABC):
                                     resource: Text,
                                     read_only: bool = False):
         pod, container = resource.split(':')
-        command = ["tar", "czf", "-", "-C", "/", posixpath.relpath(src, '/')]
+        command = ["tar", "chf", "-", "-C", "/", posixpath.relpath(src, '/')]
         response = await self.client_ws.connect_get_namespaced_pod_exec(
             name=pod,
             namespace=self.namespace or 'default',
@@ -153,7 +201,9 @@ class BaseHelmConnector(BaseConnector, ABC):
                         tar_buffer.write(data)
             await response.close()
             tar_buffer.seek(0)
-            with tarfile.open(fileobj=tar_buffer, mode='r|gz') as tar:
+            with tarfile.open(
+                    fileobj=tar_buffer,
+                    mode='r|') as tar:
                 utils.extract_tar_stream(tar, src, dst)
 
     async def _get_container(self, resource: Text) -> Tuple[Text, V1Container]:
@@ -208,10 +258,12 @@ class BaseHelmConnector(BaseConnector, ABC):
     async def deploy(self, external: bool):
         # Init standard client
         configuration = await self._get_configuration()
+        configuration.connection_pool_maxsize = self.maxConcurrentConnections
         self.client = client.CoreV1Api(api_client=ApiClient(configuration=configuration))
         # Init WebSocket client
         configuration = await self._get_configuration()
-        ws_api_client = WsApiClient(configuration=configuration)
+        configuration.connection_pool_maxsize = self.maxConcurrentConnections
+        ws_api_client = PatchedWsApiClient(configuration=configuration)
         ws_api_client.set_default_header('Connection', 'upgrade,keep-alive')
         self.client_ws = client.CoreV1Api(api_client=ws_api_client)
 
@@ -348,8 +400,7 @@ class Helm2Connector(BaseHelmConnector):
                  chartVersion: Optional[Text] = None,
                  wait: Optional[bool] = True,
                  purge: Optional[bool] = True,
-                 transferBufferSize: int = (2 ** 25) - 1,
-                 readBufferSize: Optional[int] = None):
+                 transferBufferSize: int = (2 ** 25) - 1):
         super().__init__(
             streamflow_config_dir=streamflow_config_dir,
             inCluster=inCluster,
@@ -357,7 +408,6 @@ class Helm2Connector(BaseHelmConnector):
             namespace=namespace,
             releaseName=releaseName,
             resourcesCacheTTL=resourcesCacheTTL,
-            readBufferSize=readBufferSize,
             transferBufferSize=transferBufferSize
         )
         self.chart = os.path.join(streamflow_config_dir, chart)
@@ -577,8 +627,7 @@ class Helm3Connector(BaseHelmConnector):
                  verify: Optional[bool] = False,
                  chartVersion: Optional[Text] = None,
                  wait: Optional[bool] = True,
-                 transferBufferSize: int = (32 << 20) - 1,
-                 readBufferSize: Optional[int] = None):
+                 transferBufferSize: int = (32 << 20) - 1):
         super().__init__(
             streamflow_config_dir=streamflow_config_dir,
             inCluster=inCluster,
@@ -586,7 +635,6 @@ class Helm3Connector(BaseHelmConnector):
             namespace=namespace,
             releaseName=releaseName,
             resourcesCacheTTL=resourcesCacheTTL,
-            readBufferSize=readBufferSize,
             transferBufferSize=transferBufferSize
         )
         self.chart = os.path.join(streamflow_config_dir, chart)
